
/* Router solicitation program.  Standalone.
   Does not require link-layer indications */


/*************************************************************/
/*
Please note that either this license MUST be displayed
or a particular instance of this code may be converted to
the GNU General Public License v2 or greater.

Please see README.txt for details.
*/

/*
Copyright (c) 2005, Monash University
All rights reserved.
Contributors: Greg Daley

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.

* Neither the name of Monash University nor the names of its
  contributors may be used to endorse or promote products derived
  from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO NTLINK SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
 
*/
/*************************************************************/

/* output */
#include <stdio.h>

/* time information */
#include <time.h>
#include <sys/time.h>

/* fd reading */
#include <unistd.h>
#include <sys/select.h>

#include <errno.h>

#include <bits/types.h>

/* memset */
#include <string.h>

/* free */
#include <stdlib.h>

#include <netinet/icmp6.h>



#include "fdarray.h"
#include "constants.h"
#include "prefixlist.h"
#include "inputs.h"
#include "dna.h"


extern int send_rs(int sock, int ifindex, uint8_t macaddr[], int tentative,
                   struct in6_addr *src, struct in6_addr *dst, uint8_t *
                   prefix);


#define ANCILLEN 1024

#define empty_prefixlist( P )  \
		((P)->head->next == (P)->tail) 


extern int errno;
/**************************************************************************/

/* get the ifindex as well as the data : From RFC3542 and mailing list
examples */
int getmessage(int sock, int *ifindex, unsigned char *buf, int max){
	int n;
	struct msghdr msg;
	struct cmsghdr *cmsgptr; 
	struct in6_pktinfo *pktinfo; 
	unsigned char cmbuf[ANCILLEN];
	struct iovec iov;

	/* Thanks to RFC 3542 */

	memset(&msg, 0, sizeof(msg));
	memset(&cmbuf, 0, ANCILLEN);

	iov.iov_base = buf;
	iov.iov_len = max;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

	msg.msg_control=cmbuf;
	msg.msg_controllen=ANCILLEN;	

	n = recvmsg(sock, &msg, 0);

	if (n < 0){
		return -1;
	}

	for (cmsgptr = CMSG_FIRSTHDR(&msg); cmsgptr != NULL;
				cmsgptr = CMSG_NXTHDR(&msg, cmsgptr)) {
		if (cmsgptr->cmsg_len == 0) {
			/* Error handling */
			break;
		}
		if ((cmsgptr->cmsg_level == IPPROTO_IPV6) &&
				(cmsgptr->cmsg_type == IPV6_PKTINFO)) {

			pktinfo = (struct in6_pktinfo *)CMSG_DATA(cmsgptr);

			/* process data pointed to by ptr */
			*ifindex = pktinfo->ipi6_ifindex;
			break;  /* GD ?? */
		}
	}

	return n;
}


int collect_prefixes_from_pio(struct prefixlist *pfxlist,
					struct nd_opt_prefix_info *pio){
	struct in6_addr *p;
	struct timeval expiry;

	p=&pio->nd_opt_pi_prefix;
	printf("collect_prefixes_from_pio: %p\n", pfxlist);
	printf("pfx "
		"%4.4x:%4.4x:%4.4x:%4.4x/%d\n",
		ntohs(p->s6_addr16[0]), ntohs(p->s6_addr16[1]),
		ntohs(p->s6_addr16[2]), ntohs(p->s6_addr16[3]),
		pio->nd_opt_pi_prefix_len);

	
	gettimeofday(&expiry,NULL);
	expiry.tv_sec += MIN2(pio->nd_opt_pi_preferred_time,
					MAX_DNA_PREFIX_LIFETIME); 
	insert_prefix(p, pio->nd_opt_pi_prefix_len, &expiry, 1,  pfxlist);
	print_prefixlist(pfxlist);
	return 0;

}

int collect_prefixes_from_dnao(struct prefixlist *pfxlist,
					struct nd_opt_dna_hdr *dnao, int len){
	unsigned char *dptr;
	unsigned dnaoprefs;
	unsigned dnaopoff; 
	unsigned dnaolenoff; 
	struct timeval expiry; 
	struct in6_addr *p;

	dptr=(unsigned char *)dnao;

	dnaoprefs= ((dnao->nd_opt_dna_len<<3)-2)/(1+sizeof(struct in6_addr));

	dnaolenoff = 2;

	gettimeofday(&expiry,NULL);
	expiry.tv_sec += MAX_DNA_PREFIX_LIFETIME;

	for(dnaopoff = ((dnaoprefs+9)>>3)<<3; dnaopoff < (dnao->nd_opt_dna_len
		<<3); dnaopoff += sizeof(struct in6_addr), dnaolenoff++){
		p=(struct in6_addr *)&dptr[dnaopoff];
		fprintf(stderr, "DNA opt (pfs:%d) l(%d) po(%d) o[lo](%d) "
			"%4.4x:%4.4x:%4.4x:%4.4x\n",
			dnaoprefs, dnao->nd_opt_dna_len, dnaopoff,
			dptr[dnaolenoff],
			ntohs(p->s6_addr16[0]), ntohs(p->s6_addr16[1]),
			ntohs(p->s6_addr16[2]), ntohs(p->s6_addr16[3]));
			insert_prefix(p, dptr[dnaolenoff], &expiry, 0, pfxlist);
			/*print_prefixlist(pfxlist);*/
	}

	return 0;

}


int check_landmark(struct link_status *state, struct nd_opt_landmark *lmarko){
	if(!prefix_cmp(&lmarko->nd_opt_landmark_prefix, &state->landmark,
			lmarko->nd_opt_landmark_plen, state->landplen)){
		/* have same prefix */
		if(lmarko->nd_opt_landmark_flags & ND_OPT_LMARK_FLAG_YES){
			state->received_landmark=1;
		}
		else if(lmarko->nd_opt_landmark_flags & ND_OPT_LMARK_FLAG_NO){
			state->received_landmark=-1;
		}
		fprintf(stderr, "checked landmark  %d\n",
						state->received_landmark);
	}
	return state->received_landmark; 	

}


int process_raoptions(unsigned char *buf, int n, struct link_status *state,
			 struct prefixlist *pfxlist){
	int l; 
	struct nd_opt_hdr *opt_h;
	struct nd_router_advert *ra;
	struct prefixlist store;
	struct prefixlist *candpl;
	int rec;
	/* check for completeness */

	l=sizeof(struct nd_router_advert);

	ra = (struct nd_router_advert *)buf;

	printf("process_raoptions:  pfxlist:%p\n",pfxlist);

	print_prefixlist(pfxlist);
	candpl=pfxlist; 

	if(ra->nd_ra_flags_reserved & ND_RA_FLAG_COMPLETERA){
		/* supercede existing information */
		initprefixlist(&store);
		candpl=&store;
		candpl->complete=1;
		fprintf(stderr,"RA received is COMPLETE\n");
	}

	while((l < n) ){
		opt_h = (struct nd_opt_hdr *)&buf[l];

		fprintf(stderr,"Option %d\n",opt_h->nd_opt_type);
		switch(opt_h->nd_opt_type){

			case ND_OPT_PREFIX_INFORMATION:
				
				collect_prefixes_from_pio(candpl,
					(struct nd_opt_prefix_info *)opt_h);
				break;
			case ND_OPT_DNA:
				/* awaiting testbed DNA Options */
				collect_prefixes_from_dnao(candpl,
					(struct nd_opt_dna_hdr *)opt_h,
					opt_h->nd_opt_len<<3);
				break;
			case ND_OPT_LANDMARK:
					/* only check once */
				if(state->asked_landmark && (rec=
					check_landmark(state,
					 (struct nd_opt_landmark*)opt_h))){
					if(state->require_landmark){
						state->reply_landmark=rec;
						state->require_landmark=0;
					}
				}
			case 0:
				return -1;
				break;
			default:
				break;
		}
		l += opt_h->nd_opt_len << 3;
	}
	if(ra->nd_ra_flags_reserved & ND_RA_FLAG_COMPLETERA){
		/* supercede existing information */
		/* doesn't take into account existing PIO lidetimed data */
		give_prefixlist(pfxlist, candpl);
		fprintf(stderr,"RA received is COMPLETE\n");
	}
	printf("resultant prefixlist:\n");
	print_prefixlist(pfxlist);
	return 0;
}





	/* attempt to get a function which can be placed	*/
	/* in the function pointer array 			*/
	/* Not currently used					*/
int empty_fn(int fd, struct link_status *state, struct prefixlist *lp,
							struct prefixlist *p){
	fprintf(stderr,"empty_fn\n");
	return 0;
}

	/* attempt to get a function which can be placed	*/
	/* in the function pointer array 			*/
	/* Not currently used					*/
int discard_ra(int fd, struct link_status *state, struct prefixlist *lp,
							struct prefixlist *p){
	unsigned char buf[BUFPKTSZ];
	int recvifi; 
	int n;

	while((n=getmessage(fd, &recvifi, buf, BUFPKTSZ))>0){
		fprintf(stderr, "message received on %d\n", fd);
		if(recvifi == state->ifindex){
			fprintf(stderr, "on same interface %d (%d)\n", fd,						state->ifindex);
		}
	}
	fprintf(stderr, "discard_ra\n");

	if ((n < 0 )&& ( errno != EAGAIN))
		return -1;
	return 0;
}


int process_ra( int fd, struct link_status *state, struct prefixlist *lp,
				struct prefixlist *p){
	unsigned char buf[BUFPKTSZ];
	int ret=0; 
	int recvifi;
	int n;



	while((n=getmessage(fd, &recvifi, buf, BUFPKTSZ))>0){
		if(recvifi==state->ifindex){
			fprintf(stderr, "process_ra on interesting"
				" interface %d (%d)\n", fd, state->ifindex);
			/*initprefixlist(&pfxlist);*/
			process_raoptions(buf, n, state, p);
			/* have collected prefixes */
			
			ret = 1;
		}
	}

	if ((n < 0 )&& ( errno != EAGAIN))
		return -1;

	return ret;
}


int changeinference(struct link_status *s, struct prefixlist *past, 
		struct prefixlist *recvd){
	struct prefixlist store;
	int change=0;

	/* need to be able to infer change from state and	*/
	/* gathered prefixes					*/
		
	initprefixlist(&store);
	printf("Outstanding lmark? %d (%d)\n", s->asked_landmark,
                                                       s->reply_landmark);

	printf("Any prefixes received? %d\n", !empty_prefixlist(recvd));
	print_prefixlist(recvd);

	if(s->reply_landmark){

		/* in here, transfer prefix information to link list */

		clone_prefixlist(&store,recvd);

		if(s->reply_landmark > 0 ){ /* Yes */
			/* same link, merge prefix sets */
			printf("Landmark: Y - Same link\n");
			union_prefixlist(past,&store);
			print_prefixlist(past);
		}
		else{ /* No */
			printf("Landmark: N - Different link\n");
			give_prefixlist(past, &store);	
			print_prefixlist(past);
		}
		change=(s->reply_landmark < 0);

                s->reply_landmark=0;
                s->require_landmark=0;
		s->received_landmark=0;
		s->hint_received=0;
        }
	else if(s->new_hint_received){
		printf("new hint received \n");
		s->hint_received=1;
		s->new_hint_received=0;
		clear_prefixlist(recvd);
		printf("about to send_rs %d\n",
                        s->v6sock);
                s->hint_received=1;

                s->require_landmark=1;       /* faked momentarily */
                s->asked_landmark=1;         /* faked momentarily */
                s->reply_landmark=0;
                s->received_landmark=0;
                s->tv.tv_sec=RTR_SOLICITATION_INTERVAL; s->tv.tv_usec=0;
                s->rs_transmissions=1;

                s->landplen = select_landmark(&s->landmark, past);
                send_rs(s->v6sock, s->ifindex, s->macaddr, s->tentative,
			&s->source, NULL, (uint8_t *)&s->landmark);
		change=0;
	}
	else if(empty_prefixlist(recvd)){

		return 0;
	}
	else if(does_intersect(past, recvd)){
		/* we're on the same link */
		printf("Intersecting prefix lists!\n");

		if(recvd->complete){
			supercede_prefixlist(past,recvd);
		}
		else{
			union_prefixlist(past, recvd);
		}
		print_prefixlist(past);
		s->hint_received=0;
		change=0;
	}
	else if(!s->hint_received){
		/* add prefixes if there's no hint */
		printf("no hint: add prefixes\n");
		clone_prefixlist(&store,recvd);
		union_prefixlist(past,&store);
		print_prefixlist(past);
		s->hint_received=0;
		s->require_landmark=0;
	}
	else if (past->complete){
		printf("CPL detects link change!\n");
		clone_prefixlist(&store,recvd);
		give_prefixlist(past, &store);	
		print_prefixlist(past);
		s->hint_received=0;
		change=1;
	}
	else if(s->hint_received && !empty_prefixlist(past)){
		/* we have prefixes, and haven't changed links,
			necessarily....??? */
		printf("Disjoint incomplete\n");
	}
	else{
		printf("REACHING HERE\n");
	}
	if(change && (!past->complete)){
		/* change with incompleteRA */
                printf("change with incompleteRA\n");

		s->gather_prefixes=1;
                s->tv.tv_sec=RTR_SOLICITATION_INTERVAL; /* Remaining time us?*/ 
                s->rs_transmissions=0;

                s->landplen = select_landmark(&s->landmark, past);
		return change;
	}

	return change;
}


int inputcontrol_l(int fdarray[], struct link_status *state,
			int (*funcarray[])(int, struct link_status *, 
						struct prefixlist *,
						struct prefixlist *)){
	int i;
        int maxsock=-1;
	int listenstart; 
        int numfds;
	int going=1;
	int change=0;
	int ret;
	fd_set rfds, efds;
	struct in6_addr undef = { {  { 0 } } };
	struct timeval tv;
	struct timeval now;
	struct prefixlist linkplist;
	struct prefixlist recvplist;

	initprefixlist(&linkplist);
	initprefixlist(&recvplist);
	/* initially for our test, we may have no link ??? */
	linkplist.complete=0;
	
	state->hint_received=1;

	if(memcmp(&state->landmark, &undef, sizeof(struct in6_addr))){
		state->asked_landmark=1;
		state->require_landmark=1;
	}
    /***********************************************************************/
	listenstart = SOCKET_INDEX_ICMPV6;

	maxsock=-1;
	for(i=listenstart; i <= SOCKET_INDEX_MAX; i++  ){
		if(fdarray[i] > maxsock)
			maxsock=fdarray[i];
	}


	FD_ZERO(&rfds);
	FD_ZERO(&efds);

	for(i=listenstart; i <= SOCKET_INDEX_MAX; i++  ){
		FD_SET(fdarray[i], &rfds);
        }

	tv.tv_sec=RESPONSE_SECONDS_WAIT; tv.tv_usec=RESPONSE_U_SECONDS_WAIT;

	while(going &&
		 ((numfds = select(maxsock+1,  &rfds, NULL, &efds, &tv)) > 0)){
		fprintf(stderr, "select exits (%d)\n", numfds);

		age_prefixlist(&linkplist);
		age_prefixlist(&recvplist);
		while (going && (numfds >0)){ /* O(N^2) initially! */
			fprintf(stderr, "numfds processing (%d)\n", numfds);
			for(i=listenstart; i <= SOCKET_INDEX_MAX; i++  ){
				if(FD_ISSET(fdarray[i], &rfds)){
					fprintf(stderr, "Socket activity on "
						"%d t(%lu.%6.6lu)\n", fdarray[i]
						,tv.tv_sec, tv.tv_usec);
					ret=funcarray[i](fdarray[i], state,
								&linkplist,
								&recvplist);
					if(ret >0){/* TEMP */
						change=changeinference(state,
								&linkplist,
								&recvplist);
					}
					FD_CLR(fdarray[i], &rfds);
					numfds--;
					break;
				}
				if(FD_ISSET(fdarray[i], &efds)){
					fprintf(stderr,
						"Socket error on %d\n",
						fdarray[i]);
					FD_CLR(fdarray[i], &rfds);
					going=0;
					numfds--;
					break;
				}
			}
		}
		if(change){
			fprintf(stderr,"-----------------------------------\n");
			
			gettimeofday(&now,NULL);
			fprintf(stderr," New link  (%lu.%6.6lu)\n",
				now.tv_sec, now.tv_usec);
				

			fprintf(stderr,"-----------------------------------\n");
		}

		/* all readable FDs are cleared by now */

		FD_ZERO(&rfds);
		FD_ZERO(&efds);
		for(i=listenstart; i <= SOCKET_INDEX_MAX; i++  ){
			FD_SET(fdarray[i] ,&rfds);
        	}
        } 
	return 0;
}


