

/* Router solicitation program.  Standalone.
   Does not require link-layer indications */


/*************************************************************/
/*
Please note that either this license MUST be displayed
or a particular instance of this code may be converted to
the GNU General Public License v2 or greater.

Please see README.txt for details.
*/

/*
Copyright (c) 2005, Monash University
All rights reserved.
Contributors: Greg Daley

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above
  copyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.

* Neither the name of Monash University nor the names of its
  contributors may be used to endorse or promote products derived
  from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO NTLINK SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
 
*/
/*************************************************************/
#ifndef   EVENTRS_PREFIXLIST_H
#define   EVENTRS_PREFIXLIST_H 1

#include <netinet/in.h>


/**************************************************************************/

struct prefix_e{
	struct prefix_e *next;
	struct prefix_e *prev;
	int plen;
	int recvd_pio; 
	struct timeval expiry;
	struct in6_addr prefix;
};

struct prefixlist{
	int complete;
	int rssends;
	int count;
	struct timeval lastrs;
	struct prefix_e h;
	struct prefix_e t;
	struct prefix_e *head;
	struct prefix_e *tail;
};



#define empty_prefixlist( P )  \
                ((P)->head->next == (P)->tail)


extern void prefix_copy(struct in6_addr *dest, struct in6_addr *src  , int len);
extern int prefix_cmp(struct in6_addr *p1, struct in6_addr *p2, int l1, int l2);

extern int insert_prefix(struct in6_addr *p, int l, struct timeval *exp,
				int pio, struct prefixlist *pfxlist);

extern void print_prefixlist(struct prefixlist *p);
extern void age_prefixlist(struct prefixlist *p);
extern void clone_prefixlist(struct prefixlist *to, struct prefixlist *from);
extern void clear_prefixlist(struct prefixlist *p);
extern void give_prefixlist(struct prefixlist *to, struct prefixlist *from);
extern void initprefixlist(struct prefixlist *p);
extern void supercede_prefixlist(struct prefixlist *replaced,
						struct prefixlist *from);
extern void union_prefixlist(struct prefixlist *unito, struct prefixlist *from);
extern int does_intersect(struct prefixlist *pl1, struct prefixlist *pl2);

extern int select_landmark(struct in6_addr *newland, struct prefixlist *pl);


#endif /*  EVENTRS_PREFIXLIST_H */
